// Copyright (C) 2009 by Thomas Moulard, AIST, CNRS, INRIA.
//
// This file is part of the roboptim.
//
// roboptim is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// roboptim is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with roboptim.  If not, see <http://www.gnu.org/licenses/>.

#ifndef ROBOPTIM_CORE_PROBLEM_HXX
# define ROBOPTIM_CORE_PROBLEM_HXX
# include <algorithm>
# include <stdexcept>

# include <boost/format.hpp>
# include <boost/variant.hpp>
# include <boost/variant/get.hpp>
# include <boost/variant/apply_visitor.hpp>

# include <roboptim/core/indent.hh>
# include <roboptim/core/terminal-color.hh>
# include <roboptim/core/util.hh>
# include <roboptim/core/detail/utility.hh>

namespace roboptim
{

  //
  // Template specialization for problem without constraint
  //
  template <typename F>
  Problem<F, boost::mpl::vector <> >::Problem (const function_t& f)
    : function_ (f),
      startingPoint_ (),
      argumentBounds_ (static_cast<size_t> (f.inputSize ())),
      argumentScaling_ (static_cast<size_t> (f.inputSize ())),
      argumentNames_ ()
  {
    // Check that in the objective function m = 1 (R^n -> R).
    assert (f.outputSize () == 1);

    // Initialize bound.
    std::fill (argumentBounds_.begin (), argumentBounds_.end (),
               function_t::makeInfiniteInterval ());
    // Initialize scaling.
    std::fill (argumentScaling_.begin (), argumentScaling_.end (), 1.);
  }

  // Copy constructor.
  template <typename F>
  Problem<F, boost::mpl::vector <> >::Problem
  (const Problem<F, boost::mpl::vector <> >& pb) :
    function_ (pb.function_),
    startingPoint_ (pb.startingPoint_),
    argumentBounds_ (pb.argumentBounds_),
    argumentScaling_ (pb.argumentScaling_),
    argumentNames_ (pb.argumentNames_)
  {
  }

  // Copy constructor (convert from another class of problem).
  template <typename F>
  template <typename F_>
  Problem<F, boost::mpl::vector<> >::Problem
  (const Problem<F_, boost::mpl::vector<> >& pb)
    : function_ (pb.function_),
      startingPoint_ (pb.startingPoint_),
      argumentBounds_ (pb.argumentBounds_),
      argumentScaling_ (pb.argumentScaling_),
      argumentNames_ (pb.argumentNames_)
  {
    // Check that F is a subtype of F_.
    BOOST_MPL_ASSERT_MSG((boost::is_base_of<F, F_>::value),
                         SAME_COST_FUNCTION_TYPE_EXPECTED, (F&, F_&));
  }

  template <typename F>
  Problem<F, boost::mpl::vector<> >::~Problem ()
  {
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::function_t&
  Problem<F, boost::mpl::vector<> >::function () const
  {
    return function_;
  }

  template <typename F>
  typename Problem<F, boost::mpl::vector<> >::startingPoint_t&
  Problem<F, boost::mpl::vector<> >::startingPoint ()
  {
    if (startingPoint_ && startingPoint_->size ()
	!= this->function ().inputSize ())
      assert (0 && "Invalid starting point (wrong size)");
    return startingPoint_;
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::startingPoint_t&
  Problem<F, boost::mpl::vector<> >::startingPoint () const
  {
    if (startingPoint_ && startingPoint_->size ()
	!= this->function ().inputSize ())
      assert (0 && "Invalid starting point (wrong size)");
    return startingPoint_;
  }

  template <typename F>
  std::ostream&
  operator<< (std::ostream& o,
	      const Problem<F, boost::mpl::vector<> > & pb)
  {
    return pb.print (o);
  }

  template <typename F>
  std::ostream&
  Problem<F, boost::mpl::vector<> >::
  print (std::ostream& o) const
  {
    using namespace boost;

    o << "Problem:" << incendl;
    // Function.
    o << this->function () << iendl;

    // Arguments bounds.
    o << "Arguments bounds: " << this->argumentBounds () << iendl;
    // Arguments scaling.
    o << "Arguments scaling: " << this->argumentScaling () << iendl;
    // Arguments names.
    if (!this->argumentNames ().empty ())
      o << "Arguments names: " << this->argumentNames () << iendl;

    // Starting point.
    if (startingPoint_)
      {
	o << iendl << "Starting point: "
	  << "[" << startingPoint_->size () << "](";
	for (typename F::vector_t::Index i = 0;
	     i < startingPoint_->size (); ++i)
	  {
	    if (i > 0)
	      o << ",";
	    std::size_t i_ = static_cast<std::size_t> (i);
	    if (F::getLowerBound
		(this->argumentBounds ()[i_]) <= (*startingPoint_)[i] &&
		(*startingPoint_)[i] <= F::getUpperBound
		(this->argumentBounds ()[i_]))
	      o << fg::ok << (*startingPoint_)[i] << fg::reset;
	    else
	      o << fg::fail << (*startingPoint_)[i] << fg::reset;
	  }
	o << ")" << iendl << "Starting value: "
	  << this->function () (*startingPoint_);
      }
    else
      o << iendl << fg::warn << "No starting point." << fg::reset;

    // Infinity.
    o << iendl << "Infinity value (for all functions): "
      << function_t::infinity ();
    return o << decindent;
  }

  template <typename F>
  typename Problem<F, boost::mpl::vector<> >::intervals_t&
  Problem<F, boost::mpl::vector<> >::argumentBounds ()
  {
    return argumentBounds_;
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::intervals_t&
  Problem<F, boost::mpl::vector<> >::argumentBounds () const
  {
    return argumentBounds_;
  }

  template <typename F>
  typename Problem<F, boost::mpl::vector<> >::scaling_t&
  Problem<F, boost::mpl::vector<> >::argumentScaling ()
  {
    return argumentScaling_;
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::scaling_t&
  Problem<F, boost::mpl::vector<> >::argumentScaling () const
  {
    return argumentScaling_;
  }

  template <typename F>
  typename Problem<F, boost::mpl::vector<> >::scales_t&
  Problem<F, boost::mpl::vector<> >::argumentScales ()
  {
    return argumentScaling ();
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::scales_t&
  Problem<F, boost::mpl::vector<> >::argumentScales () const
  {
    return argumentScaling ();
  }

  template <typename F>
  typename Problem<F, boost::mpl::vector<> >::names_t&
  Problem<F, boost::mpl::vector<> >::argumentNames ()
  {
    return argumentNames_;
  }

  template <typename F>
  const typename Problem<F, boost::mpl::vector<> >::names_t&
  Problem<F, boost::mpl::vector<> >::argumentNames () const
  {
    return argumentNames_;
  }

  //
  //
  // General template implementation
  //
  template <typename F, typename CLIST>
  Problem<F, CLIST>::Problem (const function_t& f)
    : function_ (f),
      startingPoint_ (),
      constraints_ (),
      boundsVect_ (),
      argumentBounds_ (static_cast<std::size_t> (f.inputSize ())),
      scalingVect_ (),
      argumentScaling_ (static_cast<std::size_t> (f.inputSize ())),
      argumentNames_ ()
  {
    // Initialize bound.
    std::fill (argumentBounds_.begin (), argumentBounds_.end (),
               function_t::makeInfiniteInterval ());
    // Initialize scaling.
    std::fill (argumentScaling_.begin (), argumentScaling_.end (), 1.);
  }

  template <typename F, typename CLIST>
  Problem<F, CLIST>::~Problem ()
  {
  }

  // Copy constructor.
  template <typename F, typename CLIST>
  Problem<F, CLIST>::Problem (const Problem<F, CLIST>& pb)
    : function_ (pb.function_),
      startingPoint_ (pb.startingPoint_),
      constraints_ (pb.constraints_),
      boundsVect_ (pb.boundsVect_),
      argumentBounds_ (pb.argumentBounds_),
      scalingVect_ (pb.scalingVect_),
      argumentScaling_ (pb.argumentScaling_),
      argumentNames_ (pb.argumentNames_)
  {
  }

  // Copy constructor (convert from another class of problem).
  template <typename F, typename CLIST>
  template <typename F_, typename CLIST_>
  Problem<F, CLIST>::Problem (const Problem<F_, CLIST_>& pb)
    : function_ (pb.function_),
      startingPoint_ (pb.startingPoint_),
      constraints_ (),
      boundsVect_ (pb.boundsVect_),
      argumentBounds_ (pb.argumentBounds_),
      scalingVect_ (pb.scalingVect_),
      argumentScaling_ (pb.argumentScaling_),
      argumentNames_ (pb.argumentNames_)
  {
    using namespace boost;

    // Check that F is a subtype of F_.
    BOOST_MPL_ASSERT_MSG((is_base_of<F, F_>::value),
                         INCOMPATIBLE_TYPES_FOR_COST, (F&, F_&));

    // Check that CLIST_ is a subset of CLIST (i.e. all the functions
    // of CLIST_ derive from functions of CLIST).
    BOOST_MPL_ASSERT_MSG(
      (detail::is_compatible_list<CLIST_, CLIST>::type::value),
       INCOMPATIBLE_TYPES_IN_LIST, (CLIST_, CLIST));

    // Copy constraints.
    constraints_.reserve (pb.constraints_.size ());
    detail::ConvertConstraintVariant<Problem<F, CLIST> > converter;
    for (typename Problem<F_, CLIST_>::constraints_t::const_iterator
	   iter  = pb.constraints_.begin ();
         iter != pb.constraints_.end ();
         ++iter)
      constraints_.push_back (apply_visitor (converter, *iter));
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::function_t&
  Problem<F, CLIST>::function () const
  {
    return function_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::constraints_t&
  Problem<F, CLIST>::constraints () const
  {
    return constraints_;
  }

// Verify that C is in CLIST or derives from one of its elements.
#define ASSERT_CONSTRAINT_TYPE(C,CLIST)                     \
  BOOST_MPL_ASSERT_MSG(                                     \
    (detail::check_constraint_type<C, CLIST>::type::value), \
     CONSTRAINT_TYPE_IS_NOT_VALID, (C&, CLIST))


  template <typename F, typename CLIST>
  template <typename C>
  void
  Problem<F, CLIST>::addConstraint (boost::shared_ptr<C> x,
				    interval_t b,
				    value_type s)
  {
    // Check that C is in CLIST.
    ASSERT_CONSTRAINT_TYPE (C, CLIST);

    // Find the proper constraint type for upcasting.
    typedef typename detail::cast_constraint_type<C, CLIST>::type
      constraintCast_t;

    if (x->inputSize () != this->function ().inputSize ())
      throw std::runtime_error ("Invalid constraint (wrong input size)");
    if (x->outputSize () != 1)
      throw std::runtime_error
	("Invalid constraint (output size is not equal to one)");

    // Check that the pointer is not null.
    assert (!!x.get ());
    assert (b.first <= b.second);
    constraints_.push_back (boost::static_pointer_cast<constraintCast_t> (x));
    intervals_t bounds;
    bounds.push_back (b);
    boundsVect_.push_back (bounds);
    scaling_t scaling;
    scaling.push_back (s);
    scalingVect_.push_back (scaling);
  }

  template <typename F, typename CLIST>
  template <typename C>
  void
  Problem<F, CLIST>::addConstraint (boost::shared_ptr<C> x,
				    intervals_t b,
				    scaling_t s)
  {
    // Check that C is in CLIST.
    ASSERT_CONSTRAINT_TYPE (C, CLIST);

    // Find the proper constraint type for upcasting.
    typedef typename detail::cast_constraint_type<C, CLIST>::type
      constraintCast_t;

    if (!x)
      throw std::runtime_error
	("Failed to add constraint: null shared pointer");

    if (x->inputSize () != this->function ().inputSize ())
      {
	boost::format fmt
	  ("Failed to add constraint '%s': invalid input size "
	   "(%d, expected size is %d)");
	fmt
	  % x->getName ()
	  % x->inputSize ()
	  % this->function ().inputSize ();
	throw std::runtime_error (fmt.str ());
      }
    if (x->outputSize () != static_cast<typename function_t::size_type> (b.size ()))
      {
	boost::format fmt
	  ("Failed to add constraint '%s': interval vector size is invalid "
	   "(%d, expected size is %d)");
	fmt
	  % x->getName ()
	  % s.size ()
	  % x->outputSize ();
	throw std::runtime_error (fmt.str ());
      }
    if (x->outputSize () != static_cast<typename function_t::size_type> (s.size ()))
      {
	boost::format fmt
	  ("Failed to add constraint '%s': scaling vector size is invalid "
	   "(%d, expected size is %d)");
	// Ignore some irrelevant exceptions
	fmt.exceptions (boost::io::all_error_bits ^ (boost::io::too_many_args_bit
						     | boost::io::too_few_args_bit
						     | boost::io::bad_format_string_bit));
	fmt
	  % x->getName ()
	  % b.size ()
	  % x->outputSize ();
	throw std::runtime_error (fmt.str ());
      }

    // Check that the pointer is not null.
    assert (!!x.get ());
    constraints_.push_back (boost::static_pointer_cast<constraintCast_t> (x));

    // Check that the bounds are correctly defined.
    for (std::size_t i = 0; i < static_cast<std::size_t> (x->outputSize ());
         ++i)
      {
        ROBOPTIM_DEBUG_ONLY(const interval_t& interval = b[i]);
	assert (interval.first <= interval.second);
      }

    boundsVect_.push_back (b);
    scalingVect_.push_back (s);
  }

#undef ASSERT_CONSTRAINT_TYPE

  template <typename F, typename CLIST>
  typename Problem<F, CLIST>::startingPoint_t&
  Problem<F, CLIST>::startingPoint ()
  {
    if (startingPoint_ && startingPoint_->size ()
	!= this->function ().inputSize ())
      throw std::runtime_error ("Invalid starting point (wrong size)");
    return startingPoint_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::startingPoint_t&
  Problem<F, CLIST>::startingPoint () const
  {
    if (startingPoint_ && startingPoint_->size ()
	!= this->function ().inputSize ())
      throw std::runtime_error ("Invalid starting point (wrong size)");
    return startingPoint_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::intervalsVect_t&
  Problem<F, CLIST>::boundsVector () const
  {
    return boundsVect_;
  }

  template <typename F, typename CLIST>
  typename Problem<F, CLIST>::intervals_t&
  Problem<F, CLIST>::argumentBounds ()
  {
    return argumentBounds_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::intervals_t&
  Problem<F, CLIST>::argumentBounds () const
  {
    return argumentBounds_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::scalingVect_t&
  Problem<F, CLIST>::scalingVector () const
  {
    return scalingVect_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::scalesVect_t&
  Problem<F, CLIST>::scalesVector () const
  {
    return scalingVector ();
  }

  template <typename F, typename CLIST>
  typename Problem<F, CLIST>::scaling_t&
  Problem<F, CLIST>::argumentScaling ()
  {
    return argumentScaling_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::scaling_t&
  Problem<F, CLIST>::argumentScaling () const
  {
    return argumentScaling_;
  }

  template <typename F, typename CLIST>
  typename Problem<F, CLIST>::scales_t&
  Problem<F, CLIST>::argumentScales ()
  {
    return argumentScaling ();
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::scales_t&
  Problem<F, CLIST>::argumentScales () const
  {
    return argumentScaling ();
  }

  template <typename F, typename CLIST>
  typename Problem<F, CLIST>::names_t&
  Problem<F, CLIST>::argumentNames ()
  {
    return argumentNames_;
  }

  template <typename F, typename CLIST>
  const typename Problem<F, CLIST>::names_t&
  Problem<F, CLIST>::argumentNames () const
  {
    return argumentNames_;
  }

  namespace detail
  {
    template <typename T>
    std::ostream&
    impl_print (std::ostream& o, Eigen::Ref<T> t)
    {
      return o << t;
    }

    template <typename T>
    std::ostream&
    impl_print (std::ostream& o, const T* t)
    {
      return o << *t;
    }

    template <typename T>
    std::ostream&
    impl_print (std::ostream& o, const T& t)
    {
      return o << t;
    }
  }

  namespace detail
  {
    template <typename P>
    struct printConstraint : public boost::static_visitor<void>
    {
      printConstraint (std::ostream& o,
		       const P& problem,
		       std::size_t i) :
	problem_ (problem),
	o_ (o),
	i_ (i)
      {}

      template <typename U>
      void operator () (const U& constraint)
      {
	assert (problem_.constraints ().size () - i_ > 0);
	using namespace boost;
        o_ << iendl << incindent
           << "Constraint " << i_ << incindent
           << iendl << *constraint
           << iendl << "Bounds: " << problem_.boundsVector ()[i_]
           << iendl << "Scaling: " << problem_.scalingVector ()[i_];

	if (problem_.startingPoint ())
	  {
	    typename P::function_t::argument_t x0 = *problem_.startingPoint ();
	    U g = get<U> (problem_.constraints ()[i_]);
            typename P::function_t::result_t x = (*g) (x0);
	    bool satisfied = true;
            o_ << iendl << "Initial value: ";
	    o_ << "[" << x.size () << "](";
            for (typename P::size_type j = 0; j < x.size (); ++j)
	      {
                if (x[j] < P::function_t::
		    getLowerBound ((problem_.boundsVector ()
				    [i_])
				   [static_cast<std::size_t> (j)])
                    || x[j] > P::function_t::
		    getUpperBound ((problem_.boundsVector ()
				    [i_])
				   [static_cast<std::size_t> (j)]))
		  {
		    satisfied = false;
		    o_ << fg::fail << x[j];
		  }
		else
		  o_ << fg::ok << x[j];

		o_ << fg::reset;
		if (j < x.size () - 1)
		  o_ << ", ";
	      }
	    o_ << ")";
	    if (!satisfied)
	      o_ << " (constraint not satisfied)";
            o_ << fg::reset;
	  }
        o_ << decindent << decindent;
      }
    private:
      const P& problem_;
      std::ostream& o_;
      std::size_t i_;
    };
  } // end of namespace detail.

  template <typename F, typename CLIST>
  std::ostream&
  Problem<F, CLIST>::print (std::ostream& o) const
  {
    using namespace boost;

    o << "Problem:" << incendl;
    // Function.
    o << this->function () << iendl;

    // Arguments bounds.
    o << "Arguments bounds: " << this->argumentBounds () << iendl;
    // Arguments scaling.
    o << "Arguments scaling: " << this->argumentScaling () << iendl;
    // Arguments names.
    if (!this->argumentNames ().empty ())
      o << "Arguments names: " << this->argumentNames () << iendl;

    // Constraints.
    if (this->constraints ().empty ())
      o << fg::ok << "No constraints." << fg::reset;
    else
      o << "Number of constraints: " << this->constraints ().size ();

    for (unsigned i = 0; i < this->constraints ().size (); ++i)
      {
	detail::printConstraint<Problem<F, CLIST> > pc (o, *this, i);
	try
	  {
	    boost::apply_visitor (pc, this->constraints ()[i]);
	  }
	catch (const boost::bad_get& e)
	  {
	    o << "failed to print constraints (boost::bad_get)" << iendl;
	  }
      }

    // Starting point.
    if (startingPoint_)
      {
	o << iendl << "Starting point: "
	  << "[" << startingPoint_->size () << "](";
	for (typename F::vector_t::Index i = 0;
	     i < startingPoint_->size (); ++i)
	  {
	    if (i > 0)
	      o << ",";
	    std::size_t i_ = static_cast<std::size_t> (i);
	    if (F::getLowerBound
		(this->argumentBounds ()[i_]) <= (*startingPoint_)[i] &&
		(*startingPoint_)[i] <= F::getUpperBound
		(this->argumentBounds ()[i_]))
	      o << fg::ok << (*startingPoint_)[i];
	    else
	      o << fg::fail << (*startingPoint_)[i];
	    o << fg::reset;
	  }
	typename F::argument_t x0 = *startingPoint_;
	o << ")" << iendl << "Starting value: "
	  << this->function () (x0);
      }
    else
      o << iendl << fg::warn << "No starting point." << fg::reset;

    // Infinity.
    o << iendl << "Infinity value (for all functions): "
      << function_t::infinity ();
    return o << decindent;
  }

  template <typename F, typename CLIST>
  std::ostream&
  operator<< (std::ostream& o, const Problem<F, CLIST>& pb)
  {
    return pb.print (o);
  }
} // end of namespace roboptim
#endif //! ROBOPTIM_CORE_PROBLEM_HH
